#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <time.h>

#define PERMS (S_IRUSR | S_IWUSR)

int iniciarSem(key_t clave, int valor, sig_atomic_t *listo) {
  int semid;
  struct timespec tiempo;
 
  semid = semget(clave, 1, PERMS | IPC_CREAT | IPC_EXCL);
  if(semid == -1 && errno != EEXIST) {  /* Error de verdad */
    perror("semid");
    return -1;
  }
  if(semid >= 0) {  /* Este proceso ha creado el semaforo: lo inicializa */
    semctl(semid, 0, SETVAL, valor);
    *listo = 1;
    return semid;
  }
  else /* El semaforo ya existe: simplemente accedemos a el */
    if((semid = semget(clave, 1, PERMS)) == -1)
      return -1;
  /* Esperar a que se inicialice el semaforo */

  tiempo.tv_sec = 0;
  tiempo.tv_nsec = 10000000;
  while(*listo == 0)
    nanosleep(&tiempo, NULL); 

  return semid;
}

int main() {
  char c;
  key_t k;
  int shid, semid;
  sig_atomic_t *listo;
  struct sembuf esperar[1], avisar[1];

  /* Inicializar memoria compartida */
  k = ftok("abece.c", 4);
  shid = shmget(k, sizeof(sig_atomic_t), PERMS | IPC_CREAT | IPC_EXCL);
  if((shid == -1) && (errno != EEXIST))   /* Error real */
    return -1;
  if(shid == -1) {  /* Ya se ha creado; acceder y adjuntarla */
    shid = shmget(k, sizeof(sig_atomic_t), PERMS);
    if(shid == -1) {
      perror("Al acceder a memoria compartida");
      return -1;
    }
    listo = (sig_atomic_t *) shmat(shid, NULL, 0);
  }
  else {  /* Se ha creado con exito; adjuntar e inicializar */
    listo = (sig_atomic_t *) shmat(shid, NULL, 0);
    fprintf(stderr, "inicializo\n");
    *listo = 0;
  }

  /* Crear (o acceder al) semaforo */
  semid = iniciarSem(k, 1, listo);

  /* Inicializar variables para modificar el semaforo */
  esperar[0].sem_num = 0;
  esperar[0].sem_op = -1;
  esperar[0].sem_flg = 0;

  avisar[0].sem_num = 0;
  avisar[0].sem_op = 1;
  avisar[0].sem_flg = 0;

  /* Entrar en la seccion critica */
  semop(semid, esperar, 1);

  for(c = 'a'; c <= 'z'; c++)
    fprintf(stderr, "%c ", c);
  fprintf(stderr, "\n");

  /* Salir de la seccion critica */
  semop(semid, avisar, 1);

  return 0;
}

