#include <stdio.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/stat.h>

main(int argc, char *argv[])
{
  int *buf;
  int i;
  int shmid, semid, llave;
  struct sembuf operacion;

  if((shmid = shmget(IPC_PRIVATE, sizeof(int) * 5, S_IRUSR | S_IWUSR)) == -1)
    {
      perror("shmget");
      exit(-1);
    }

  buf = (int *) shmat(shmid, 0, 0);

  /* Se crean dos semaforos: el primero lleva los huecos libre y el otro los 
     huecos ocupados. Se podria resolver el problema utilizando solo uno,  
     pero asi resulta mas claro */
  if((semid = semget(IPC_PRIVATE, 2, S_IRUSR | S_IWUSR)) == -1)
    {
      perror("semget");
      exit(-1);
    }
  
  semctl(semid, 0, SETVAL, 5);
  semctl(semid, 1, SETVAL, 0);

  switch(fork()) {
    case -1:
      perror("fork");
      exit(-1);
    case 0:
      for(i = 1; i <= 200; i++)
	{
	  /* Comprueba que hay sitio para escribir */
	  operacion.sem_num = 0;
	  operacion.sem_op = -1;
	  operacion.sem_flg = 0;
	  semop(semid, &operacion, 1);

	  buf[(i-1) % 5] = i;
	  printf("Hijo: %d\n", i);

          /* Actualizar numero de elementos disponibles */
	  operacion.sem_num = 1;
	  operacion.sem_op = 1;
	  operacion.sem_flg = 0;
	  semop(semid, &operacion, 1);
	}

      shmdt(buf);
      exit(0);
    default:
      for(i = 1; i <= 200; i++)
	{
	  /* Comprueba que hay datos que leer */
	  operacion.sem_num = 1;
	  operacion.sem_op = -1;
	  operacion.sem_flg = 0;
	  semop(semid, &operacion, 1);

          /* Con un unico semaforo, se podria resolver mediante una espera
             activa:
             
             while(semctl(semid, 0, GETVAL, 0) == 5);
	  */

	  fprintf(stderr, "Padre: %d\n", buf[(i-1) % 5]);

          /* Actualizar numero de huecos libres */
	  operacion.sem_num = 0;
	  operacion.sem_op = 1;
	  operacion.sem_flg = 0;
	  semop(semid, &operacion, 1);
	}

      semctl(semid, 0, IPC_RMID, 0);
      shmdt(buf);
      shmctl(shmid, IPC_RMID, NULL);
      exit(0);
    }
}
